__author__ = "Felix Kleinert"
__data__ = "2021-02-03"

import xarray as xr
import numpy as np
import pandas as pd
import itertools
import glob
import matplotlib.pyplot as plt
from dask.diagnostics import ProgressBar

import dask
import inspect
import os
import gc
from mlair.helpers.geofunctions import haversine_dist, bearing_angle, WindSector, VectorRotateLambertConformal2latlon
from mlair.helpers.helpers import convert2xrda, remove_items, to_list
from mlair.helpers import TimeTrackingWrapper, TimeTracking


from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler import DefaultDataHandler

from typing import Tuple, Union, Dict, List
import logging

float_np_xr = Union[float, np.ndarray, xr.DataArray, xr.Dataset]
int_or_list_of_int = Union[int, List[int]]


class BaseWrfChemDataLoader:
    DEFAULT_LOGICAL_TIME_COORD_NAME = 'Time'
    DEFAULT_LOGICAL_X_COORD_NAME = 'west_east'
    DEFAULT_LOGICAL_Y_COORD_NAME = 'south_north'
    DEFAULT_LOGICAL_Z_COORD_NAME = 'bottom_top'

    DEFAULT_PHYSICAL_TIME_COORD_NAME = "XTIME"
    DEFAULT_PHYSICAL_X_COORD_NAME = 'XLONG'
    DEFAULT_PHYSICAL_Y_COORD_NAME = 'XLAT'

    DEFAULT_RECHUNK = {"Time": -1, "y": 36, "x": 40}
    DEFAULT_FILE_STARTER = 'wrfout_d0'

    DEFAULT_STAGED_VARS = ('U', 'V')
    DEFAULT_STAGED_ROTATION_opts = dict(cen_lon=12., cen_lat=52.5,
                                        truelat1=30., truelat2=60.,
                                        stand_lon=12.)
    DEFAULT_VARS_TO_ROTATE = ((('U', 'V'), ('Ull', 'Vll')),
                              (('U10', 'V10'), ('U10ll', 'V10ll')),
                              (('U', 'V'), ('wspdll', 'wdirll')),
                              (('U10', 'V10'), ('wspd10ll', 'wdir10ll'))
                              )

    def __init__(self, data_path: str, common_file_starter: str = DEFAULT_FILE_STARTER,
                 time_dim_name: str = DEFAULT_LOGICAL_TIME_COORD_NAME,
                 rechunk_values: Dict = None,
                 logical_x_coord_name: str = DEFAULT_LOGICAL_X_COORD_NAME,
                 logical_y_coord_name: str = DEFAULT_LOGICAL_Y_COORD_NAME,
                 logical_z_coord_name: str = DEFAULT_LOGICAL_Z_COORD_NAME,
                 physical_x_coord_name: str = DEFAULT_PHYSICAL_X_COORD_NAME,
                 physical_y_coord_name: str = DEFAULT_PHYSICAL_Y_COORD_NAME,
                 physical_t_coord_name: str = DEFAULT_PHYSICAL_TIME_COORD_NAME,
                 staged_vars: List[str] = DEFAULT_STAGED_VARS,
                 variables=None, z_coord_selector=None,
                 start_time=None, end_time=None,
                 staged_rotation_opts: Dict = DEFAULT_STAGED_ROTATION_opts,
                 vars_to_rotate: Tuple[Tuple[Tuple[str, str], Tuple[str, str]]] = DEFAULT_VARS_TO_ROTATE,
                 staged_dimension_mapping=None, stag_ending='_stag',
                 date_format_of_nc_file=None,

                 ):
        # super().__init__()
        self.data_path = data_path
        self.common_file_starter = common_file_starter
        self.time_dim_name = time_dim_name

        self.logical_x_coord_name = logical_x_coord_name
        self.logical_y_coord_name = logical_y_coord_name
        self.logical_z_coord_name = logical_z_coord_name

        self.physical_x_coord_name = physical_x_coord_name
        self.physical_y_coord_name = physical_y_coord_name
        self.physical_t_coord_name = physical_t_coord_name

        self.staged_rotation_opts = staged_rotation_opts
        self.vars_to_rotate = vars_to_rotate

        self.variables = variables
        self.z_coord_selector = z_coord_selector

        self.start_time = start_time
        self.end_time = end_time
        if date_format_of_nc_file is None:
            date_format_of_nc_file = "%Y-%m-%d"
        self.date_format_of_nc_file = date_format_of_nc_file

        # if rechunk_values is None:
        #     self.rechunk_values = {self.time_dim_name: 1}
        # else:
        #     self.rechunk_values = rechunk_values
        self.rechunk_values = rechunk_values

        self._stag_ending = stag_ending
        if staged_dimension_mapping is None:
            self.staged_dimension_mapping = {'X': self.logical_x_coord_name+self._stag_ending,
                                             'Y': self.logical_y_coord_name+self._stag_ending,
                                             'Z': self.logical_z_coord_name+self._stag_ending,}
        else:
            self.staged_dimension_mapping = staged_dimension_mapping

        # internal
        self._X = None
        self._Y = None
        self._data = None

    @property
    def staged_dimension_mapping(self):
        return self.__staged_dimension_mapping

    @staged_dimension_mapping.setter
    def staged_dimension_mapping(self, var: dict):
        if not isinstance(var, dict):
            raise TypeError(f"`var' must be a dict, but is of type {type(var)}")
        self.__staged_dimension_mapping = var

    @property
    def data(self) -> xr.Dataset:
        return self._data

    @property
    def dataset_search_str(self):
        if (self.start_time is None) and (self.end_time is None):
            path_list = os.path.join(self.data_path, self.common_file_starter + '*')
            logging.info(f"Reading file(s): {path_list}")
            return path_list
        elif (self.start_time is not None) and (self.end_time is not None):
            path_list = set()
            for day in pd.date_range(self.start_time, self.end_time):
                path_list.add(sorted(glob.glob(os.path.join(self.data_path,
                                                            self.common_file_starter) + '*' + day.strftime(
                    self.date_format_of_nc_file) + '*'
                                               ))[0])
            path_list = sorted(list(path_list))
            logging.info(f"Reading file(s): {path_list}")
            return path_list
        else:
            raise ValueError(f"`start_time' and `end_time' must both be given or None.")

    @TimeTrackingWrapper
    def open_data(self):
        # see also https://github.com/pydata/xarray/issues/1385#issuecomment-438870575
        # data = xr.open_mfdataset(paths=self.dataset_search_str, combine='nested', concat_dim=self.time_dim_name,
        #                          parallel=True, decode_cf=False)
        if self.variables is None:
            # see also https://github.com/pydata/xarray/issues/1385#issuecomment-438870575
            data = xr.open_mfdataset(paths=self.dataset_search_str, combine='nested', concat_dim=self.time_dim_name,
                                     parallel=True, decode_cf=False)
        else:
            data = xr.open_mfdataset(paths=self.dataset_search_str, combine='nested', concat_dim=self.time_dim_name,
                                     parallel=True, decode_cf=False, preprocess=self.preprocess_fkt_for_loader, )
        data = xr.decode_cf(data)
        # if self.variables is not None:
        #     data = self.preprocess_fkt_for_loader(data)

        # if self.rechunk_values is None:
        #     chunks = {k: 'auto' for k in data.chunks.keys() }
        #     chunks[self.time_dim_name] = -1
        #     data = data.chunk(chunks)
        #     # data = data.chunk("auto")
        # else:
        #     data = data.chunk(self.rechunk_values)
        self._data = data

    def preprocess_fkt_for_loader(self, ds):
        # ToDo make genreal: Currently it's in fact hardcoded!
        wind_var_mapping = {'Ull': ['U'], 'Vll': ['V'], 'U10ll': ['U10'], 'V10ll': ['V10'],
                            'wspdll': ['U', 'V'], 'wdirll': ['U', 'V'],
                            'wspd10ll': ['U10', 'V10'], 'wdir10ll': ['U10', 'V10'],
                            }
        potential_wind_vars_list = list(
            set(itertools.chain(
                *itertools.chain(*SingleGridColumnWrfChemDataLoader.DEFAULT_VARS_TO_ROTATE))))
        none_wind_vars_to_keep = [x for x in self.variables if x not in potential_wind_vars_list]
        wind_vars_to_keep = ['U', 'V', 'U10', 'V10']
        combined_vars_to_keep = none_wind_vars_to_keep + wind_vars_to_keep
        ds = ds[combined_vars_to_keep]
        if self.z_coord_selector is not None:
            ds = ds.sel({self.logical_z_coord_name: self.z_coord_selector})
        return ds

    def assign_coords(self, coords, **coords_kwargs):
        """
        Assign coords to WrfChemDataHandler._data

        """
        self._data = self._data.assign_coords(coords, **coords_kwargs)

    def rechunk_data(self, chunks=None, name_prefix='xarray-', token=None, lock=False):
        return self._data.chunk(chunks=chunks, name_prefix=name_prefix, token=token, lock=lock)

    def read_filenames(self):
        combs = []
        start = 1
        end = 9
        for l in range(1, end):
            combs.extend(itertools.combinations(range(start, end), l))

        for i in combs:
            filenamestr = f'{self.dataset_search_str[:-1]}{list(i)}_*'
            try:
                with xr.open_mfdataset(paths=filenamestr, combine='nested', concat_dim='Time',
                                       parallel=True, chunks={'Time': 12}) as data:
                    status = 'OK'
            except:
                status = 'FAIL'
            print(f'{i}: {filenamestr} {status}')

    def get_distances(self, lat, lon):
        dist = haversine_dist(lat1=self._data.XLAT, lon1=self._data.XLONG, lat2=lat, lon2=lon)
        return dist

    def get_bearing(self, lat, lon, points_last=True):
        bearing = bearing_angle(lat1=lat, lon1=lon, lat2=self._data.XLAT, lon2=self._data.XLONG)
        if points_last:
            return bearing.transpose(..., 'points')
        else:
            return bearing

    def compute_nearest_icoordinates(self, lat, lon, dim=None):
        dist = self.get_distances(lat=lat, lon=lon)

        if dim is None:
            return dist.argmin(...)
        else:
            return dist.argmin(dim)

    def _set_dims_as_coords(self):
        if self._data is None:
            raise IOError(f'{self.__class__.__name__} can not set dims as coords. Use must use `open_data()` before.')
        data = self._data
        for k, _ in data.dims.items():
            data = data.assign_coords({k: data[k]})
        self._data = data
        # logging.info('set dimensions as coordinates')

    def apply_staged_transormation(self, mapping_of_stag2unstag=None):
        if mapping_of_stag2unstag is None:
            mapping_of_stag2unstag = {'U': 'U10', 'V': 'V10', 'U10': 'U10', 'V10': 'V10'}
        vectorrot = VectorRotateLambertConformal2latlon(xlat=self._data[self.physical_y_coord_name].data,
                                                        xlong=self._data[self.physical_x_coord_name].data,
                                                        **self.staged_rotation_opts)
        if self._data is None:
            raise IOError(f"`open_data' must be called before vector interpolation and rotation")
        for (u_stag_name, v_stag_name), (u_ll_name, v_ll_name) in self.vars_to_rotate:
            u_staged_field = self._data[u_stag_name]
            v_staged_field = self._data[v_stag_name]

            u_target_field = self._data[mapping_of_stag2unstag[u_stag_name]]
            v_target_field = self._data[mapping_of_stag2unstag[v_stag_name]]

            # This if statements are needed as cdo might have removed empty attrs from variables
            if "stagger" not in u_staged_field.attrs:
                u_staged_field.attrs['stagger'] = ""
            if "stagger" not in v_staged_field.attrs:
                v_staged_field.attrs['stagger'] = ""

            u_stagger = u_staged_field.attrs['stagger']
            v_stagger = v_staged_field.attrs['stagger']

            if u_stagger:
                u_grd = self.set_interpolated_field(staged_field=u_staged_field, target_field=u_target_field)
            else:
                u_grd = self._data[u_stag_name]

            if v_stagger:
                v_grd = self.set_interpolated_field(staged_field=v_staged_field, target_field=v_target_field)
            else:
                v_grd = self._data[v_stag_name]

            if u_ll_name[0] == 'U' and v_ll_name[0] == 'V':
                ull, vll = vectorrot.ugrd_vgrd2ull_vll(u_grd, v_grd)
            elif u_ll_name[:4] == 'wspd' and v_ll_name[:4] == 'wdir':
                ull, vll = vectorrot.ugrd_vgrd2wspd_wdir(u_grd, v_grd)
            else:
                raise ValueError(
                    f"`u_ll_name' and `u_ll_name' must either start with 'U' and 'V' or 'wspd' and 'wdir', "
                    f"but they are u_ll_name={u_ll_name} and v_ll_name={v_ll_name}")

            self._data[u_ll_name] = ull
            self._data[v_ll_name] = vll

    def set_interpolated_field(self, staged_field: xr.DataArray, target_field: xr.DataArray,
                               dropped_staged_attrs: List[str] =None, **kwargs):
        stagger_attr_name = kwargs.pop('stagger_attr_name', 'stagger')
        stagger = kwargs.pop('stagger', staged_field.attrs[stagger_attr_name])
        if dropped_staged_attrs is None:
            dropped_staged_attrs = [stagger_attr_name]

        excluded_dims = [self.staged_dimension_mapping[stagger]]
        target_dim_order = [dim.replace('_stag', '') for dim in staged_field.dims]
        new_field = xr.broadcast(target_field, staged_field, exclude=excluded_dims)[0]
        new_field = new_field.transpose(*target_dim_order)
        new_field.attrs.update(remove_items(staged_field.attrs, dropped_staged_attrs))

        new_field.data = VectorRotateLambertConformal2latlon.interpolate_to_grid_center(staged_field,
                                                                                        self.staged_dimension_mapping[stagger])
        return new_field


class SingleGridColumnWrfChemDataLoader(BaseWrfChemDataLoader):
    DEFAULT_MODEL = "WRF-Chem"
    DEFAULT_VAR_ALL_DICT = {'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                            'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                            'pblheight': 'maximum'}
    DEFAULT_WINDOW_LEAD_TIME = 3
    DEFAULT_WINDOW_HISTORY_SIZE = 13
    DEFAULT_WINDOW_HISTORY_OFFSET = 0
    DEFAULT_TARGET_VAR = "o3"
    DEFAULT_TARGET_DIM = "variables"
    DEFAULT_ITER_DIM = "points"
    DEFAULT_WINDOW_DIM = "window"

    def __init__(self, coords: Tuple[float_np_xr, float_np_xr],
                 target_dim=DEFAULT_TARGET_DIM, target_var=DEFAULT_TARGET_VAR,
                 window_history_size=DEFAULT_WINDOW_HISTORY_SIZE,
                 window_lead_time=DEFAULT_WINDOW_LEAD_TIME,
                 external_coords_file: str = None,
                 wind_sectors=None, **kwargs):
        super().__init__(**kwargs)
        self._set_coords(coords)
        self.target_dim = target_dim
        self.target_var = target_var
        self.window_history_size = window_history_size
        self.window_lead_time = window_lead_time
        self._nearest_icoords = None
        self._nearest_coords = None
        self.external_coords_file = external_coords_file
        self.wind_sectors = wind_sectors

        logging.debug("SingleGridColumnWrfChemDataLoader Initialised")

    def __enter__(self):
        self.open_data()

        if self.physical_t_coord_name != self.time_dim_name:
            self.assign_coords(
                {self.physical_t_coord_name: (self.time_dim_name, self._data[self.physical_t_coord_name].values)})

        self._set_dims_as_coords()
        if self.external_coords_file is not None:
            self._apply_external_coordinates()
        self.apply_staged_transormation()
        # self.rechunk_data(self.rechunk_values)
        self._set_geoinfos()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.data.close()
        gc.collect()

    def _set_geoinfos(self):
        # identify nearest coords
        self._set_nearest_icoords(dim=[self.logical_x_coord_name, self.logical_y_coord_name])
        self._set_nearest_coords()

        # get lat and lons
        lat = self.get_coordinates()['lat']
        lon = self.get_coordinates()['lon']

        # calculate distances to points (from points to all grid boxes in lat/lon space)
        dist = self.get_distances(lat, lon)
        dist.attrs.update(dict(units='km'))

        # calculate bearing angle
        bearing = self.get_bearing(lat, lon)
        bearing.attrs.update(dict(units='°'))

        # set geoinfo data set
        self._geo_infos = xr.Dataset({'dist': dist, 'bearing': bearing})

        # set wind_sectors
        # initialize WindSector() providing helper methods for wind sector calculations
        ws = WindSector(wind_sectors=self.wind_sectors)

        # expand dataset by new dataarray
        self._geo_infos['wind_sectors'] = xr.full_like(bearing, fill_value=np.nan) # xr.DataArray(coords=bearing.coords, dims=bearing.dims)
        for i, (k, v) in enumerate(ws.wind_sectore_edges.items()):
            self._geo_infos['wind_sectors'] = xr.where(ws.is_in_sector(k, self.geo_infos['bearing']), i, self._geo_infos['wind_sectors'])
        # self._geo_infos['wind_sectors'].attrs.update(dict(units='Wind Sector'))

    @property
    def geo_infos(self):
        return self._geo_infos

    def _apply_external_coordinates(self):
        ds_coords = xr.open_dataset(self.external_coords_file, chunks={'south_north': 36, 'west_east': 40})
        data = self._data
        for k, v in ds_coords.coords.variables.items():
            data = data.assign_coords({k: (remove_items(list(v.dims), 'Time'), v.values.squeeze())})
        self._data = data
        ds_coords.close()
        # logging.debug('setup external coords')

    def _set_coords(self, coords):
        __set_coords = dict(lat=None, lon=None)
        if len(coords) != 2:
            raise SyntaxError(f"`coords' must have length=2 (lat, lon), but has length={len(coords)}")
        if isinstance(coords, tuple):
            self.__coords = dict(lat=coords[0], lon=coords[1])
            logging.debug(f"{self.__class__.__name__}.__coords is set to {self.__coords}")
        elif isinstance(coords, dict):
            if (coords.keys() == __set_coords.keys()):
                self.__coords = coords
            else:
                raise KeyError(f"dict `coords' must have the keys `lat', and `lon' but has: {coords.keys()}")
        else:
            raise TypeError(f"`coords' must be a tuple of floats or a dict, but is of type: {type(coords)}")

    def get_coordinates(self, as_arrays=False) -> Union[Tuple[np.ndarray, np.ndarray], dict]:
        if as_arrays:
            return np.array(self.__coords['lat']), np.array(self.__coords['lon'])
        else:
            return self.__coords

    def _set_nearest_icoords(self, dim=None):
        lat, lon = self.get_coordinates(as_arrays=True)
        with ProgressBar():
            logging.info("SingleGridColumnWrfChemDataLoader: compute nearest icoordinates")
            self._nearest_icoords = dask.compute(self.compute_nearest_icoordinates(lat, lon, dim))[0]

    def get_nearest_icoords(self, as_arrays=False):
        if as_arrays:
            return (self._nearest_icoords[self.logical_y_coord_name].values,
                    self._nearest_icoords[self.logical_x_coord_name].values,
                    )
        else:
            return {k: list(v.values) for k, v in self._nearest_icoords.items()}

    def _set_nearest_coords(self):
        icoords = self.get_nearest_icoords()
        ilat = convert2xrda(np.array(icoords[self.logical_y_coord_name]), use_1d_default=True)
        ilon = convert2xrda(np.array(icoords[self.logical_x_coord_name]), use_1d_default=True)
        lat = self._data[self.physical_y_coord_name].isel(
            {self.logical_x_coord_name: ilon, self.logical_y_coord_name: ilat})
        lon = self._data[self.physical_x_coord_name].isel(
            {self.logical_x_coord_name: ilon, self.logical_y_coord_name: ilat})
        self._nearest_coords = dict(lat=lat, lon=lon)

    def get_nearest_coords(self, as_arrays=False):
        if as_arrays:
            return (self._nearest_coords['lat'].values,
                    self._nearest_coords['lon'].values,
                    )
        else:
            return {k: list(v.values) for k, v in self._nearest_coords.items()}


class DataHandlerSingleGridColumn(DataHandlerSingleStation):
    _requirements = remove_items(inspect.getfullargspec(DataHandlerSingleStation).args, ["self", "station"])

    def __init__(self, *args, external_coords_file=None,
                 var_logical_z_coord_selector=None,
                 targetvar_logical_z_coord_selector=None,
                 rechunk_values=None,
                 date_format_of_nc_file=None,
                 as_image_like_data_format=True,
                 **kwargs):
        self.external_coords_file = external_coords_file
        self.var_logical_z_coord_selector = self._ret_z_coord_select_if_valid(var_logical_z_coord_selector,
                                                                              as_input=True)
        self.targetvar_logical_z_coord_selector = self._ret_z_coord_select_if_valid(targetvar_logical_z_coord_selector,
                                                                                    as_input=False)
        self._logical_z_coord_name = None
        self._joint_z_coord_selector = self._extract_largest_coord_extractor(self.var_logical_z_coord_selector,
                                                                             self.targetvar_logical_z_coord_selector)
        self.__loader = None
        self.rechunk_values = rechunk_values
        self.date_format_of_nc_file = date_format_of_nc_file
        self.as_image_like_data_format = as_image_like_data_format
        super().__init__(*args, **kwargs)

    @staticmethod
    def _ret_z_coord_select_if_valid(z_coord: int_or_list_of_int, as_input: bool) -> int_or_list_of_int:
        if isinstance(z_coord, int):
            return z_coord
        elif isinstance(z_coord, list):
            all_entries_int = all(isinstance(zi, int) for zi in z_coord)
            if all_entries_int and as_input:
                return z_coord
            else:
                raise NotImplementedError(
                    f"z_coord selector of type list not implemented for target vars. "
                    f"`all_entries_int'={all_entries_int}, `as_input'= {as_input}")
        else:
            raise TypeError(f"`z_coord' must be of type int or list[int], but is {type(z_coord)}")

    @staticmethod
    def coord_str2coords(str_coords: str, sep='__', dec_marker='_') -> Tuple[float_np_xr, float_np_xr]:
        if isinstance(str_coords, list) and len(str_coords) == 1:
            str_coords = str_coords[0]
        _, lat, lon = str_coords.split(sep=sep)
        lat = lat.replace(dec_marker, '.')
        lon = lon.replace(dec_marker, '.')

        return np.array(float(lat)), np.array(float(lon))

    def setup_data_path(self, data_path: str, sampling: str):
        return data_path

    @TimeTrackingWrapper
    def extract_data_from_loader(self, loader):
        data = loader.data.isel(loader.get_nearest_icoords()).squeeze(dim=[loader.logical_y_coord_name,
                                                                           loader.logical_x_coord_name])[self.variables]
        with ProgressBar():
            data = dask.compute(data)[0]
        return data

    @property
    def loader(self):
        return self.__loader

    @loader.setter
    def loader(self, station_path):
        try:
            station, path = station_path
        except ValueError as e:
            raise ValueError(f"Pass an iterable with two items; (station, path)")
        lat, lon = self.coord_str2coords(station)
        loader = SingleGridColumnWrfChemDataLoader((lat, lon),
                                                   data_path=path,
                                                   external_coords_file=self.external_coords_file,
                                                   time_dim_name=self.time_dim,
                                                   rechunk_values=self.rechunk_values,
                                                   variables=self.variables,
                                                   z_coord_selector=self._joint_z_coord_selector,
                                                   # start_time=self.start,
                                                   # end_time=self.end,
                                                   date_format_of_nc_file=self.date_format_of_nc_file
                                                   )
        self.__loader = loader

    def load_data(self, path, station, statistics_per_var, sampling, station_type=None, network=None,
                  store_data_locally=False, data_origin: Dict = None, start=None, end=None):

        self.loader = (station, path)

        with self.loader as loader:
            if self._logical_z_coord_name is None:
                self._logical_z_coord_name = loader.logical_z_coord_name
            # # select defined variables at grid box or grid coloumn based on nearest icoords
            data = self.extract_data_from_loader(loader)
            if self._joint_z_coord_selector is not None:
                data = data.sel({self._logical_z_coord_name: self._joint_z_coord_selector})
            with ProgressBar(), TimeTracking(name=f"{self.station}: get data"):
                logging.info(f"start compute data for {self.station} in load_data")
                data = dask.compute(data)[0]

        # expand dimesion for iterdim
        data = data.expand_dims({self.iter_dim: station}).to_array(self.target_dim)
        # transpose dataarray: set first three fixed and keep remaining as is
        data = data.transpose(self.iter_dim, self.time_dim, self.target_dim, ...)
        data = self._slice_prep(data, start=start, end=end)
        # ToDo add metadata
        _meta = {'station_lon': self.loader.get_coordinates()['lon'].tolist(),
                 'station_lat': self.loader.get_coordinates()['lat'].tolist()}
        meta = pd.DataFrame(_meta, index=station).T

        return data, meta

    @staticmethod
    def _extract_largest_coord_extractor(var_extarctor, target_extractor) -> Union[List, None]:
        if var_extarctor is not None and target_extractor is not None:
            res = list(set(to_list(var_extarctor) + to_list(target_extractor)))
        else:
            res = None
        return res

    def get_X(self, upsampling=False, as_numpy=False):
        x_data = self.get_transposed_history()
        if self.as_image_like_data_format is False:
            x_data = x_data.squeeze()
        if as_numpy is True:
            # return None
            raise NotImplementedError(f"keyword argument `as_numpy=True' not implemented.")
        elif as_numpy is False:
            return x_data

    # def get_Y(self, upsampling=False, as_numpy=False):
    #     raise NotImplementedError

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables.
        """
        return self.history.transpose(self.time_dim, self.window_dim, self.iter_dim, self.target_dim, ...).copy()

    def get_transposed_label(self) -> xr.DataArray:
        """Return label.

        :return: label with dimensions datetime*, window*, Stations, variables.
        """
        return self.label.squeeze([self.iter_dim, self.target_dim]).transpose(self.time_dim, self.window_dim, ...).copy()

    @TimeTrackingWrapper
    def make_samples(self):
        self.make_history_window(self.target_dim, self.window_history_size, self.time_dim)
        if self.var_logical_z_coord_selector is not None:
            self.history = self.history.sel({self._logical_z_coord_name: self.var_logical_z_coord_selector})
        self.make_labels(self.target_dim, self.target_var, self.time_dim, self.window_lead_time)
        self.make_observation(self.target_dim, self.target_var, self.time_dim)
        if self.targetvar_logical_z_coord_selector is not None:
            self.label = self.label.sel({self._logical_z_coord_name: self.targetvar_logical_z_coord_selector})
            self.observation = self.observation.sel(
                {self._logical_z_coord_name: self.targetvar_logical_z_coord_selector})

        self.remove_nan(self.time_dim)

        self.history = self.modify_history()
        self.label = self.modify_label()
        self.observation = self.modify_observation()

        self.remove_nan(self.time_dim)
        # self.input_data = self.input_data.compute()
        # self.label = self.label.compute()
        # self.observation = self.observation.compute()
        # self.target_data = self.target_data.compute()
        # self._data.close()

    def modify_history(self):
        """
        Place holder for more user spec. processing of history.
        :param **kwargs:
        :type **kwargs:
        :return:
        :rtype:
        """
        return self.history

    def modify_label(self):
        """
        Place holder for more user spec. processing of label.
        :param **kwargs:
        :type **kwargs:
        :return:
        :rtype:
        """
        return self.label

    def modify_observation(self, **kwargs):
        """
        Place holder for more user spec. processing of observation.
        :param **kwargs:
        :type **kwargs:
        :return:
        :rtype:
        """
        return self.observation

    @property
    def transformation_is_applied(self):
        """
        Returns True if transformation is applied.
        Checks if all entries in `self._transform' are not None/empty
        """
        return all(self._transformation)


class DataHandlerWRF(DefaultDataHandler):
    """Data handler using DataHandlerSingleGridColumn."""
    data_handler = DataHandlerSingleGridColumn
    data_handler_transformation = DataHandlerSingleGridColumn
    _requirements = data_handler.requirements()


class DataHandlerSectorGrid(DataHandlerSingleGridColumn):
    # _requirements1 = remove_items(inspect.getfullargspec(DataHandlerWRF).args, ["self", "station"])
    # _requirements2 = remove_items(inspect.getfullargspec(DataHandlerSingleGridColumn).args, ["self", "station"])
    _requirements = DataHandlerWRF.requirements()

    def __init__(self, *args, radius=None, wind_sectors=None, wind_sector_edge_dim_name=None,  **kwargs):
        if radius is None:
            radius = 100  # km
        self.radius = radius
        if wind_sectors is None:
            wind_sectors = WindSector.DEFAULT_WIND_SECTORS
        self.wind_sectors = wind_sectors
        if wind_sector_edge_dim_name is None:
            wind_sector_edge_dim_name = 'edges'
        self.wind_sector_edge_dim_name = wind_sector_edge_dim_name
        self.windsector = WindSector(wind_sectors=self.wind_sectors,
                                     wind_sector_edge_dim_name=self.wind_sector_edge_dim_name)
        self._added_vars = []
        self.wind_dir_name = None
        super().__init__(*args, **kwargs)

    @TimeTrackingWrapper
    def extract_data_from_loader(self, loader):
        wind_dir_name = self._get_wind_dir_var_name(loader)
        full_data = loader.data.isel(loader.get_nearest_icoords()).squeeze(dim=[loader.logical_y_coord_name,
                                                                                loader.logical_x_coord_name])
        data = full_data[self.variables]
        sec_data = self.windsector.get_sect_of_value(full_data[wind_dir_name])
        self.sec_data = sec_data
        if wind_dir_name not in data:
            data[wind_dir_name] = full_data[wind_dir_name]
            self._added_vars.append(to_list(wind_dir_name))
        with ProgressBar():
            data = dask.compute(data)[0]
        return data

    def _get_wind_dir_var_name(self, loader, wdir_name3d='wdirll', wdir_name2d='wdir10ll'):
        """
        Get variable name of wind direction. If no wind variable is given in self.variables, 10m wind is used as default.
        If a wind variable is used in self.variables, returns the variable name (e.g. wdirll if Ull or Vll is in self.variables,
        or wdir10ll if U10ll or V10ll is in self.variables. The same holds if wdirll/wspdll or wdir10ll/wspd10ll is given.

        :return:
        :rtype:
        """

        if (wdir_name3d in self.variables) and not (wdir_name2d in self.variables):
            var_name = wdir_name3d
        elif ((wdir_name2d in self.variables) and not (wdir_name3d in self.variables)) or (not (wdir_name2d in self.variables) and not (wdir_name3d in self.variables)):
            var_name = wdir_name2d
        else:
            raise ValueError(f"invalid wind var name passed to `_get_wind_dir_var_name'")

        assert var_name in itertools.chain(*itertools.chain(*loader.vars_to_rotate))
        if self.wind_dir_name is None:
            self.wind_dir_name = var_name
        return var_name

    def _get_left_and_right_wind_sector_edges(self, **kwargs):
        ws_edges = self.windsector.wind_sector_edges_data(**kwargs)
        ws_edges = ws_edges.expand_dims(dim={self.target_dim: to_list(self.wind_dir_name)})
        return ws_edges

    def get_transformation_opts(self, base=0):
        pos_of_trafo = self.get_transformation_base(base)
        trafo_opts = self._transformation[pos_of_trafo]
        return trafo_opts

    def apply_transformation_on_data(self, data):
        trafo_opts = self.get_transformation_opts()
        trafo, _ = self.transform(data, dim=self.aggregation_dim, inverse=False,
                                  opts=trafo_opts, transformation_dim=self.target_dim)
        return trafo

    @TimeTrackingWrapper
    def modify_history(self):
        if self.transformation_is_applied:
            ws_edges = self.get_applied_transdormation_on_wind_sector_edges()
            wind_dir_of_interest = self.compute_wind_dir_of_interest()
            sector_allocation = self.windsector.get_sect_of_value(value=wind_dir_of_interest, external_edges=ws_edges)
            existing_sectors = np.unique(sector_allocation.data)
            sector_history = xr.ones_like(self.history)
            sector_history_var_names = [f"{var}_sect" for var in sector_history.coords[self.target_dim].values]
            with self.loader as loader, TimeTracking(name="loader in modify history"):
                # setup sector history
                grid_data = self.preselect_and_transform_neighbouring_data_based_on_radius(loader)
                for i, sect in enumerate(existing_sectors, start=1):
                    with TimeTracking(name=f"processing sector {sect} ({i}/{len(existing_sectors)}) for {self.station}"):
                        # select data in wind sector
                        sec_data = self.get_section_data_from_circle(grid_data, loader, sect)
                        sec_data = self.apply_aggregation_method_on_sector_data(sec_data, loader)
                        sec_data_history = self._make_and_return_history_window(dim_name_of_shift=self.time_dim,
                                                                                window=self.window_history_size,
                                                                                data=sec_data.to_array(self.target_dim)
                                                                                )
                        sec_data_history = sec_data_history.broadcast_like(self.history)
                        sec_data_history = sec_data_history.transpose(*self.history.dims)
                        time_index_of_sector = sector_history[self.time_dim].where((sector_allocation.squeeze() == sect),
                                                                                   drop=True).values
                        sector_history.loc[{self.time_dim: time_index_of_sector}] = sec_data_history.sel(
                            {self.time_dim: time_index_of_sector}).values

            sector_history = sector_history.assign_coords({self.target_dim: sector_history_var_names})
            combined_history = xr.concat([self.history, sector_history], dim=self.target_dim)

            return combined_history
        else:
            return self.history

    def get_section_data_from_circle(self, grid_data, loader, sect, compute=True):
        sec_data = grid_data.where(
            self.windsector.is_in_sector(sect, loader.geo_infos.bearing.drop('points').squeeze()))
        if compute:
            with ProgressBar():
                return dask.compute(sec_data)[0]
        else:
            return sec_data

    @TimeTrackingWrapper
    def preselect_and_transform_neighbouring_data_based_on_radius(self, loader, compute=True):
        """
        Select neighbouring grid boxes which have a maximal distance of pre-selected radius from full model field
        :param loader:
        :type loader:
        :return:
        :rtype:
        """
        # get data from loader
        grid_data = loader.data[self.variables]
        # select correct time steps and vertical layers
        grid_data = grid_data.sel({
            # self.time_dim: self.history.coords[self.time_dim].values,  # get data which is used in history
            self.time_dim: slice(self.start, self.end),  # get data which is used in history
            self._logical_z_coord_name: self.var_logical_z_coord_selector,  # get only used vertical layers
        })
        # select grid boxes which are closer than given radius
        grid_data = grid_data.where(
            loader.geo_infos.dist.sel({"points": 0}).drop("points") <= self.radius)
        # apply transformation on variables
        grid_data = self.apply_transformation_on_data(grid_data.to_array(self.target_dim)).to_dataset(self.target_dim)

        if compute:
            return dask.compute(grid_data)[0]
        else:
            return grid_data

    @staticmethod
    def apply_aggregation_method_on_sector_data(data, loader):
        data = data.mean(dim=(loader.logical_x_coord_name, loader.logical_y_coord_name))
        return data

    def compute_wind_dir_of_interest(self):
        wind_dir_of_intrest = self.history.sel({self.target_dim: self.wind_dir_name, self.window_dim: 0})
        return wind_dir_of_intrest

    def get_applied_transdormation_on_wind_sector_edges(self):
        ws_edges = self._get_left_and_right_wind_sector_edges(return_as='xr.da', dim=self.wind_sector_edge_dim_name)
        ws_edges = self.apply_transformation_on_data(ws_edges)
        return ws_edges


class DataHandlerMainSectWRF(DefaultDataHandler):
    """Data handler using DataHandlerSectorGrid."""
    data_handler = DataHandlerSectorGrid
    data_handler_transformation = DataHandlerSectorGrid
    _requirements = data_handler.requirements()


if __name__ == '__main__':

    import cartopy.crs as ccrs
    from shapely.geometry import Point, Polygon
    import math
    from cartopy.geodesic import Geodesic
    # ToDo
    # test imports for circle plots
    import shapely


    def plot_map_proj(data, xlim=None, ylim=None, filename=None, point=None, radius=None, **kwargs):
        cmap = kwargs.pop('cmap', plt.cm.Reds)
        set_background = kwargs.pop('set_background', False)
        edgecolors = kwargs.pop('edgecolors', 'face')
        circle = kwargs.pop('circle', False)

        crs_latlon = ccrs.PlateCarree()
        if ylim is None:
            ylim = [-90, 90]
        if xlim is None:
            xlim = [-180, 180]
        if filename is None:
            filename = 'test_fig.pdf'
        if point is not None:
            lat, lon = point
        else:
            lat, lon = None, None
        # plt.figure(figsize=(14, 6))
        plt.figure()
        ax = plt.axes(projection=ccrs.PlateCarree())
        ax.set_global()
        pm = data.squeeze().plot.pcolormesh(ax=ax, transform=ccrs.PlateCarree(), x='XLONG', y='XLAT',
                                       cmap=cmap, edgecolors=edgecolors,
                                       **kwargs)
        ax.coastlines(resolution='50m')


        ##### Idea to draw circle from https://stackoverflow.com/a/58735566
        #### own try:
        if circle and ((lon is not None) and (lat is not None)):
            circle_points = Geodesic().circle(lon=lon, lat=lat,
                                              radius=radius*1000., # radius_in_meters,
                                              )
            geom = shapely.geometry.Polygon(circle_points)
            ax.add_geometries((geom,), crs=ccrs.PlateCarree(), facecolor='none', edgecolor='blue', linewidth=1)
        #
        #####

        ax.set_ylim(ylim)
        ax.set_xlim(xlim)
        if (lon is not None) and (lat is not None):
            plt.plot(lon, lat, '+k', markersize=7, transform=crs_latlon)
        if set_background:
            ax.stock_img()
        # plt.tight_layout()
        plt.savefig(filename, bbox_inches='tight')
        plt.close('all')

    def dummy_plot(data, skipna=None, xlim=None, ylim=None):
        plt.figure()
        crs_latlon = ccrs.PlateCarree()
        ax = plt.axes(projection=ccrs.PlateCarree())
        ax.set_global()
        data.sum(dim='points',skipna=skipna).squeeze().plot.pcolormesh(ax=ax, transform=ccrs.PlateCarree(), x='XLONG', y='XLAT',
                                     cmap=plt.cm.Reds, )
        ax.coastlines(resolution='50m')
        ax.set_ylim(ylim)
        ax.set_xlim(xlim)
        plt.tight_layout()
        plt.savefig("Example_Multiple_points_test.pdf")
        print("plot multiple points")

    def sector(center, start_angle, end_angle, radius, steps=200):
        def polar_point(origin_point, angle, distance):
            return [origin_point.x + math.sin(math.radians(angle)) * distance,
                    origin_point.y + math.cos(math.radians(angle)) * distance]

        if start_angle > end_angle:
            start_angle = start_angle - 360
        else:
            pass
        step_angle_width = (end_angle - start_angle) / steps
        sector_width = (end_angle - start_angle)
        segment_vertices = []

        segment_vertices.append(polar_point(center, 0, 0))
        segment_vertices.append(polar_point(center, start_angle, radius))

        for z in range(1, steps):
            segment_vertices.append((polar_point(center, start_angle + z * step_angle_width, radius)))
        segment_vertices.append(polar_point(center, start_angle + sector_width, radius))
        segment_vertices.append(polar_point(center, 0, 0))
        return Polygon(segment_vertices)

    # lat_np = np.array([50.73333, 47.73333])
    # lon_np = np.array([7.1, 8.6])
    # 48_8479__10_0963
    lat_np = np.array([48.8479])
    lon_np = np.array([10.0963])

    lat_xr = xr.DataArray(lat_np, dims=["points"], coords={'points': range(len(lat_np))})
    lon_xr = xr.DataArray(lon_np, dims=["points"], coords={'points': range(len(lon_np))})

    use_first_dummy_dataset = True
    if use_first_dummy_dataset:

        wrf_new = SingleGridColumnWrfChemDataLoader((lat_xr, lon_xr),
                                                    # data_path='/home/felix/Data/WRF-Chem/upload_aura_2021-02-24/2009/',
                                                    data_path="/media/felix/INTENSO/WRF_CHEM/JFM_2009/",
                                                    common_file_starter='wrfout_d01_',
                                                    time_dim_name='XTIME',
                                                    logical_x_coord_name='west_east',
                                                    logical_y_coord_name='south_north',
                                                    logical_z_coord_name='bottom_top',
                                                    rechunk_values={'XTIME': 24, 'bottom_top': 10},
                                                    # external_coords_file='/home/felix/Data/WRF-Chem/upload_aura_2021-02-24/coords.nc',
                                                    external_coords_file="/media/felix/INTENSO/WRF_CHEM/monthly/coords.nc",
                                                    wind_sectors=['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW'],
                                                    )
        windsect = WindSector(wind_sectors=['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW'])
        radius_from_point = 100.  # in km

        x_de_extract = [2, 15]
        y_de_extract = [45, 60]

        x_eu_extract = [-4, 20]
        y_eu_extract = [40, 60]

        with wrf_new as loader:
            o3 = loader.data.o3

        plot_map_proj(data=o3.isel({'XTIME': 0, 'bottom_top': 0}), xlim=x_de_extract, ylim=y_de_extract,
                      filename="ozone_map_eur.pdf", cbar_kwargs={'orientation': 'horizontal', 'pad': 0.01},
                      circle=False)
        icoords = dask.compute(wrf_new.compute_nearest_icoordinates(lat_np, lon_np))[0]

        dist_np = wrf_new.get_distances(lat_np, lon_np)
        dist_xr = wrf_new.get_distances(lat_xr, lon_xr)
        dist_xr.attrs.update(dict(units='km'))

        bearing_np = wrf_new.get_bearing(lat_np, lon_np)
        bearing_xr = wrf_new.get_bearing(lat_xr, lon_xr)
        bearing_xr.attrs.update(dict(units='°'))

        geo_info = xr.Dataset({'dist': dist_xr, 'bearing': bearing_xr})


        ###test for rotation
        vecrot = VectorRotateLambertConformal2latlon(xlat=wrf_new.data.XLAT.data, xlong=wrf_new.data.XLONG.data)
        ugrd = vecrot.interpolate_to_grid_center(wrf_new.data.U, 'west_east_stag')
        vgrd = vecrot.interpolate_to_grid_center(wrf_new.data.V, 'south_north_stag')
        ull_grd, vll_grd = vecrot.ugrd_vgrd2ull_vll(ugrd, vgrd)
        ull_stg, vll_stg = vecrot.ustg_vstg2ull_vll(wrf_new.data.U, wrf_new.data.V,
                                                    ustg_dim='west_east_stag', vstg_dim='south_north_stag')

        wspd_grd, wdir_grd = vecrot.ugrd_vgrd2wspd_wdir(ugrd, vgrd)
        wspd_stg, wdir_stg = vecrot.ustg_vstg2wspd_wdir(wrf_new.data.U, wrf_new.data.V,
                                                        ustg_dim='west_east_stag', vstg_dim='south_north_stag')
        ### end test rotation

        point = [wrf_new.get_nearest_coords()['lat'][0], wrf_new.get_nearest_coords()['lon'][0]]
        for i, (data, xlim, ylim, point, kwargs) in enumerate(((wrf_new._data.o3.isel({'XTIME': 0,'bottom_top': 0}),
                                                                x_eu_extract, y_eu_extract, point,
                                                         {'circle': True,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}),
                                                         x_eu_extract, y_eu_extract, point,
                                                         {'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new.geo_infos.dist, x_eu_extract, y_eu_extract, point,
                                                         {'cmap': plt.cm.get_cmap('Greys'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new.geo_infos.bearing, x_eu_extract, y_eu_extract, point,
                                                         {'cmap': plt.cm.get_cmap('twilight'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .7}}),
                                                        (wrf_new.geo_infos.wind_sectors, x_eu_extract, y_eu_extract, point,
                                                         {
                                                            'cmap': plt.cm.get_cmap('tab20', len(wrf_new.wind_sectors)),
                                                            'circle': False,
                                                            'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}
                                                            #'set_background': True,
                                                            # 'levels': 16,
                                                           }),
                                                        (wrf_new._data.o3.isel({'XTIME': 0,'bottom_top': 0}).where(
                                                            windsect.is_in_sector('N', wrf_new.geo_infos['bearing'])),
                                                         x_eu_extract, y_eu_extract, point,
                                                         {'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01}}),
                                                        (wrf_new._data.o3.isel({'XTIME': 0,'bottom_top': 0}).where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract, y_de_extract, None,
                                                         {'circle': True,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01}}),
                                                        (geo_info.dist.where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract, y_de_extract, None,
                                                         {'cmap': plt.cm.get_cmap('Greys'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal', 'pad': 0.01,
                                                                          'shrink': .6}}),
                                                        (wrf_new.geo_infos.bearing.where(
                                                            dist_xr.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract, y_de_extract, None,
                                                         {'cmap': plt.cm.get_cmap('twilight'), 'circle': False,
                                                          'cbar_kwargs': {'orientation': 'horizontal',
                                                                          'pad': 0.01, 'shrink': .65}}),
                                                        (wrf_new.geo_infos.wind_sectors.where(
                                                            dist_xr.sel({'points': 0}).drop('points') <= radius_from_point),
                                                         x_de_extract,y_de_extract, None,
                                                         {'cmap': plt.cm.get_cmap('tab20', len(wrf_new.wind_sectors)),
                                                                    'circle': False,
                                                                    'cbar_kwargs': {'orientation': 'horizontal',
                                                                                    'shrink': .6, 'pad': 0.01,
                                                                                    }
                                                                    #'set_background': True,
                                                                    }),
                                                        (wrf_new.data.o3.isel({'XTIME': 0,'bottom_top': 0}).where(
                                                            wrf_new.geo_infos.dist.sel({'points': 0}).drop('points') <= radius_from_point).where(
                                                            windsect.is_in_sector('N', wrf_new.geo_infos['bearing'])),
                                                         x_de_extract, y_de_extract, None,
                                                         {'circle': False, 'cbar_kwargs': {'orientation': 'horizontal',
                                                                                                     'fraction':.1, 'pad': 0.01,
                                                                                                         }}),
                                                        )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          # point=[lat_np, lon_np],
                          # point=[wrf_new.get_nearest_coords()['lat'][0], wrf_new.get_nearest_coords()['lon'][0]],
                          point=point,
                          filename=f'Example_sector{i}.pdf',
                          radius=radius_from_point,
                          # cbar_kwargs={'orientation': 'horizontal'},
                          **kwargs,
                          )

        for i, (data, xlim, ylim) in enumerate(
                ((wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}), [-42, 66], [23, 80]),
                 (geo_info.dist, [-42, 66], [23, 80]),
                 (wrf_new._data.o3.isel({'XTIME': 0, 'bottom_top': 0}).where(
                     dist_xr.sel({'points': 0}).drop('points') <= radius_from_point), x_de_extract, y_de_extract),
                 (geo_info.dist.where(dist_xr.sel({'points': 0}).drop('points') <= radius_from_point), x_de_extract, y_de_extract),
                 )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          point=[lat_np, lon_np],
                          filename=f'ExampleO3_dist{i}.pdf',
                          radius=radius_from_point)

    ######################### # Larger 4D data
    use_second_dummy_dataset = False
    if use_second_dummy_dataset:
        wrf_dh_4d = BaseWrfChemDataLoader(data_path='/home/felix/Data/WRF-Chem/upload_aura/2009/',
                                          common_file_starter='wrfout_d01_2009', time_dim_name='Time')
        wrf_dh_4d.open_data()
        wrf_dh_4d.rechunk_data({"XTIME": 1, "bottom_top": 34, "south_north": 36, "west_east": 40})
        lat_np = np.array([50.73333])
        lon_np = np.array([7.1])
        wrf_dh_4d._data = wrf_dh_4d._data.assign_coords(wrf_dh_4d._data.coords)
        icoords = dask.compute(wrf_dh_4d.compute_nearest_icoordinates(lat_np, lon_np))[0]

        dist_xr = wrf_dh_4d.get_distances(lat_xr, lon_xr)
        dist_xr.attrs.update(dict(units='km'))
        geo_info = xr.Dataset({'dist': dist_xr})

        for i, (data, xlim, ylim) in enumerate(((wrf_dh_4d._data.T2, [-42, 66], [23, 80]),
                                                (geo_info.dist, [-42, 66], [23, 80]),
                                                (wrf_dh_4d._data.T2.where(
                                                    dist_xr.sel({'points': 0}).drop('points') <= 100), x_de_extract,
                                                 y_de_extract),
                                                (geo_info.dist.where(
                                                    dist_xr.sel({'points': 0}).drop('points') <= 100), x_de_extract,
                                                 y_de_extract),
                                                )):
            plot_map_proj(data, xlim=xlim,
                          ylim=ylim,
                          point=[lat_np, lon_np], filename=f'test_dist{i}.pdf')

    print()
