"""
Module for neural models to use during experiment.

To work properly, each customised model needs to inherit from AbstractModelClass and needs an implementation of the
set_model method.

In this module, you can find some exemplary model classes that have been build and were running in a experiment.

* `MyLittleModel`: small model implementation with a single 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time).
* `MyBranchedModel`: a model with single  1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), it has three
  output branches from different layers of the model.
* `MyTowerModel`: a more complex model with inception blocks (called towers)
* `MyPaperModel`: A model used for the publication: <Add Publication Title / Citation>

In addition, a short introduction how to create your own model is given hereinafter.

How to create a customised model?
#################################

* Create a new class:

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def __init__(self, input_shape: list, output_shape: list):

                super().__init__(input_shape[0], output_shape[0])

                # settings
                self.dropout_rate = 0.1
                self.activation = keras.layers.PReLU

                # apply to model
                self.set_model()
                self.set_compile_options()
                self.set_custom_objects(loss=self.compile_options['loss'])

* Make sure to add the `super().__init__()` and at least `set_model()` and `set_compile_options()` to your custom init
  method.
* If you have custom objects in your model, that are not part of keras, you need to add them to custom objects. To do
  this, call `set_custom_objects` with arbitrarily kwargs. In the shown example, the loss has been added, because it
  wasn't a standard loss. Apart from this, we always encourage you to add the loss as custom object, to prevent
  potential errors when loading an already created model instead of training a new one.
* Build your model inside `set_model()`, e.g.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_model(self):
                x_input = keras.layers.Input(shape=self._input_shape)
                x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1'.format("major"))(x_input)
                x_in = self.activation(name='{}_conv_act'.format("major"))(x_in)
                x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
                x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
                x_in = keras.layers.Dense(16, name='{}_Dense_16'.format("major"))(x_in)
                x_in = self.activation()(x_in)
                x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
                out_main = self.activation()(x_in)
                self.model = keras.Model(inputs=x_input, outputs=[out_main])

* Your are free, how to design your model. Just make sure to save it in the class attribute model.
* Additionally, set your custom compile options including the loss.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_compile_options(self):
                self.initial_lr = 1e-2
                self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
                self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr,
                                                                                       drop=.94,
                                                                                       epochs_drop=10)
                self.loss = keras.losses.mean_squared_error
                self.compile_options = {"metrics": ["mse", "mae"]}

* If you have a branched model with multiple outputs, you need either set only a single loss for all branch outputs or
  to provide the same number of loss functions considering the right order. E.g.

    .. code-block:: python

        class MyCustomisedModel(AbstractModelClass):

            def set_model(self):
                ...
                self.model = keras.Model(inputs=x_input, outputs=[out_minor_1, out_minor_2, out_main])

            def set_compile_options(self):
                self.loss = [keras.losses.mean_absolute_error] +  # for out_minor_1
                            [keras.losses.mean_squared_error] +   # for out_minor_2
                            [keras.losses.mean_squared_error]     # for out_main


How to access my customised model?
##################################

If the customised model is created, you can easily access the model with

>>> MyCustomisedModel().model
<your custom model>

The loss is accessible via

>>> MyCustomisedModel().loss
<your custom loss>

You can treat the instance of your model as instance but also as the model itself. If you call a method, that refers to
the model instead of the model instance, you can directly apply the command on the instance instead of adding the model
parameter call.

>>> MyCustomisedModel().model.compile(**kwargs) == MyCustomisedModel().compile(**kwargs)
True

"""

import mlair.model_modules.keras_extensions

__author__ = "Lukas Leufen, Felix Kleinert"
__date__ = '2020-05-12'

import keras

from mlair.model_modules import AbstractModelClass
from mlair.model_modules.inception_model import InceptionModelBase
from mlair.model_modules.flatten import flatten_tail
from mlair.model_modules.advanced_paddings import PadUtils, Padding2D, SymmetricPadding2D


class MyLittleModelHourly(AbstractModelClass):
    """
    A customised model with a 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the
    output layer depending on the window_lead_time parameter. Dropout is used between the Convolution and the first
    Dense layer.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param shape_inputs: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param shape_outputs: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.1
        self.regularizer = keras.regularizers.l2(0.001)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        """
        Build the model.
        """

        # add 1 to window_size to include current time step t0
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(128, (1, 1), padding='same', name='{}_Conv_1x1_128'.format("major"))(x_input)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(64, (1, 1), padding='same', name='{}_Conv_1x1_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1_32'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
        x_in = keras.layers.Dense(128, name='{}_Dense_128'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(64, name='{}_Dense_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
        out_main = self.activation()(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-2
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class MyBranchedModel(AbstractModelClass):
    """
    A customised model

    with a 1x1 Conv, and 4 Dense layers (64, 32, 16, window_lead_time), where the last layer is the
    output layer depending on the window_lead_time parameter. Dropout is used between the Convolution and the first
    Dense layer.
    """

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.1
        self.regularizer = keras.regularizers.l2(0.1)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"])

    def set_model(self):
        """
        Build the model.
        """

        # add 1 to window_size to include current time step t0
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(32, (1, 1), padding='same', name='{}_Conv_1x1'.format("major"))(x_input)
        x_in = self.activation(name='{}_conv_act'.format("major"))(x_in)
        x_in = keras.layers.Flatten(name='{}'.format("major"))(x_in)
        x_in = keras.layers.Dropout(self.dropout_rate, name='{}_Dropout_1'.format("major"))(x_in)
        x_in = keras.layers.Dense(64, name='{}_Dense_64'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        out_minor_1 = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("minor_1"))(x_in)
        out_minor_1 = self.activation(name="minor_1")(out_minor_1)
        x_in = keras.layers.Dense(32, name='{}_Dense_32'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        out_minor_2 = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("minor_2"))(x_in)
        out_minor_2 = self.activation(name="minor_2")(out_minor_2)
        x_in = keras.layers.Dense(16, name='{}_Dense_16'.format("major"))(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape, name='{}_Dense'.format("major"))(x_in)
        out_main = self.activation(name="main")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_minor_1, out_minor_2, out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-2
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
                                                                               epochs_drop=10)
        self.compile_options = {"loss": [keras.losses.mean_absolute_error] + [keras.losses.mean_squared_error] + [
            keras.losses.mean_squared_error], "metrics": ["mse", "mae"]}


class MyTowerModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 1e-2
        self.regularizer = keras.regularizers.l2(0.1)
        self.initial_lr = 1e-2
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr, drop=.94,
                                                                               epochs_drop=10)
        self.activation = keras.layers.PReLU

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"])

    def set_model(self):
        """
        Build the model.
        """
        activation = self.activation
        conv_settings_dict1 = {
            'tower_1': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (3, 1), 'activation': activation},
            'tower_2': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (5, 1), 'activation': activation},
            'tower_3': {'reduction_filter': 8, 'tower_filter': 8 * 2, 'tower_kernel': (1, 1), 'activation': activation},
        }

        pool_settings_dict1 = {'pool_kernel': (3, 1), 'tower_filter': 8 * 2, 'activation': activation}

        conv_settings_dict2 = {
            'tower_1': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 8 * 2, 'tower_filter': 16 * 2 * 2, 'tower_kernel': (1, 1),
                        'activation': activation},
        }
        pool_settings_dict2 = {'pool_kernel': (3, 1), 'tower_filter': 16, 'activation': activation}

        conv_settings_dict3 = {'tower_1': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (3, 1),
                                           'activation': activation},
                               'tower_2': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (5, 1),
                                           'activation': activation},
                               'tower_3': {'reduction_filter': 16 * 4, 'tower_filter': 32 * 2, 'tower_kernel': (1, 1),
                                           'activation': activation},
                               }

        pool_settings_dict3 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        ##########################################
        inception_model = InceptionModelBase()

        X_input = keras.layers.Input(shape=self._input_shape)

        X_in = inception_model.inception_block(X_input, conv_settings_dict1, pool_settings_dict1,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict2, pool_settings_dict2,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict3, pool_settings_dict3,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True)
        #############################################

        # out_main = flatten_tail(X_in, 'Main', activation=activation, bound_weight=True, dropout_rate=self.dropout_rate,
        #                         reduction_filter=64, inner_neurons=64, output_neurons=self.window_lead_time)

        out_main = flatten_tail(X_in, inner_neurons=64, activation=activation, output_neurons=self._output_shape,
                                output_activation='linear', reduction_filter=64,
                                name='Main', bound_weight=True, dropout_rate=self.dropout_rate,
                                kernel_regularizer=self.regularizer
                                )

        self.model = keras.Model(inputs=X_input, outputs=[out_main])

    def set_compile_options(self):
        self.optimizer = keras.optimizers.adam(lr=self.initial_lr)
        self.compile_options = {"loss": [keras.losses.mean_squared_error], "metrics": ["mse"]}


class MyPaperModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):
        """
        Sets model and loss depending on the given arguments.

        :param input_shape: list of input shapes (expect len=1 with shape=(window_hist, station, variables))
        :param output_shape: list of output shapes (expect len=1 with shape=(window_forecast))
        """

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        from mlair.model_modules.keras_extensions import LearningRateDecay

        # settings
        self.dropout_rate = .3
        self.regularizer = keras.regularizers.l2(0.001)
        self.initial_lr = 1e-3
        self.lr_decay = LearningRateDecay(base_lr=self.initial_lr, drop=.94, epochs_drop=10)
        self.activation = keras.layers.ELU
        self.padding = "SymPad2D"

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"],
                                SymmetricPadding2D=SymmetricPadding2D,
                                LearningRateDecay=LearningRateDecay)

    def set_model(self):
        """
        Build the model.

        :param activation: activation function
        :param window_history_size: number of historical time steps included in the input data
        :param channels: number of variables used in input data
        :param dropout_rate: dropout rate used in the model [0, 1)
        :param window_lead_time: number of time steps to forecast in the output layer
        :return: built keras model
        """
        activation = self.activation
        first_kernel = (3, 1)
        first_filters = 16

        conv_settings_dict1 = {
            'tower_1': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            # 'tower_2': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (5, 1),
            #             'activation': activation},
            # 'tower_3': {'reduction_filter': 8, 'tower_filter': 16 * 2, 'tower_kernel': (1, 1),
            #             'activation': activation},
            # 'tower_4':{'reduction_filter':8, 'tower_filter':8*2, 'tower_kernel':(7,1), 'activation':activation},
        }
        pool_settings_dict1 = {'pool_kernel': (3, 1), 'tower_filter': 16, 'activation': activation}

        conv_settings_dict2 = {
            'tower_1': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (3, 1),
                        'activation': activation},
            # 'tower_2': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (5, 1),
            #             'activation': activation},
            # 'tower_3': {'reduction_filter': 64, 'tower_filter': 32 * 2, 'tower_kernel': (1, 1),
            #             'activation': activation},
            # 'tower_4':{'reduction_filter':8*2, 'tower_filter':16*2, 'tower_kernel':(7,1), 'activation':activation},
        }
        pool_settings_dict2 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        conv_settings_dict3 = {
            'tower_1': {'reduction_filter': 64 * 2, 'tower_filter': 32 * 4, 'tower_kernel': (3, 1),
                        'activation': activation},
            'tower_2': {'reduction_filter': 64 * 2, 'tower_filter': 32 * 4, 'tower_kernel': (5, 1),
                        'activation': activation},
            'tower_3': {'reduction_filter': 64 * 2, 'tower_filter': 32 * 4, 'tower_kernel': (1, 1),
                        'activation': activation},
            # 'tower_4':{'reduction_filter':16*4, 'tower_filter':32, 'tower_kernel':(7,1), 'activation':activation},
        }
        pool_settings_dict3 = {'pool_kernel': (3, 1), 'tower_filter': 32, 'activation': activation}

        ##########################################
        inception_model = InceptionModelBase()

        X_input = keras.layers.Input(shape=self._input_shape)

        pad_size = PadUtils.get_padding_for_same(first_kernel)
        X_in = Padding2D("SymPad2D")(padding=pad_size, name="SymPad")(X_input)
        X_in = keras.layers.Conv2D(filters=first_filters,
                                   kernel_size=first_kernel,
                                   kernel_regularizer=self.regularizer,
                                   name="First_conv_{}x{}".format(first_kernel[0], first_kernel[1]))(X_in)
        X_in = self.activation(name='FirstAct')(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict1, pool_settings_dict1,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True,
                                               padding=self.padding)
        # out_minor1 = flatten_tail(X_in, 'minor_1', False, self.dropout_rate, self.window_lead_time,
        #                           self.activation, 32, 64)
        out_minor1 = flatten_tail(X_in, inner_neurons=64, activation=activation, output_neurons=self._output_shape,
                                  output_activation='linear', reduction_filter=32,
                                  name='minor_1', bound_weight=False, dropout_rate=self.dropout_rate,
                                  kernel_regularizer=self.regularizer
                                  )

        X_in = keras.layers.Dropout(self.dropout_rate)(X_in)

        X_in = inception_model.inception_block(X_in, conv_settings_dict2, pool_settings_dict2,
                                               regularizer=self.regularizer,
                                               batch_normalisation=True, padding=self.padding)

        # X_in = keras.layers.Dropout(self.dropout_rate)(X_in)
        #
        # X_in = inception_model.inception_block(X_in, conv_settings_dict3, pool_settings_dict3, regularizer=self.regularizer,
        #                                        batch_normalisation=True)
        #############################################

        out_main = flatten_tail(X_in, inner_neurons=64 * 2, activation=activation, output_neurons=self._output_shape,
                                output_activation='linear', reduction_filter=64 * 2,
                                name='Main', bound_weight=False, dropout_rate=self.dropout_rate,
                                kernel_regularizer=self.regularizer
                                )

        self.model = keras.Model(inputs=X_input, outputs=[out_minor1, out_main])

    def set_compile_options(self):
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.compile_options = {"loss": [keras.losses.mean_squared_error, keras.losses.mean_squared_error],
                                "metrics": ['mse', 'mae']}


class MyLSTMModel(AbstractModelClass):

    def __init__(self, input_shape: list, output_shape: list):

        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.dropout_rate = 0.25

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options['loss'])

    def set_model(self):
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.LSTM(16, return_sequences=True, name="First_LSTM", dropout=self.dropout_rate)(x_input)
        x_in = keras.layers.LSTM(32, name="Second_LSTM", dropout=self.dropout_rate)(x_in)
        out_main = keras.layers.Dense(self._output_shape, name='Output_Dense')(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out_main])

    def set_compile_options(self):
        self.initial_lr = 1e-4
        self.optimizer = keras.optimizers.SGD(lr=self.initial_lr, momentum=0.9)
        self.lr_decay = mlair.model_modules.keras_extensions.LearningRateDecay(base_lr=self.initial_lr,
                                                                               drop=.94,
                                                                               epochs_drop=10)
        self.loss = keras.losses.mean_squared_error
        self.compile_options = {"metrics": ["mse", "mae"]}
